package com.jiuqi.util.csv;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.OutputStream;
import java.io.Writer;
import java.nio.charset.Charset;
import java.io.PrintWriter;

public class CsvWriter
{
    private PrintWriter outputStream;
    private String fileName;
    private boolean firstColumn;
    private boolean useCustomRecordDelimiter;
    private Charset charset;
    private UserSettings userSettings;
    private boolean initialized;
    private boolean closed;
    public static final int ESCAPE_MODE_DOUBLED = 1;
    public static final int ESCAPE_MODE_BACKSLASH = 2;
    
    public CsvWriter(final String fileName, final char delimiter, final Charset charset) {
        this.outputStream = null;
        this.fileName = null;
        this.firstColumn = true;
        this.useCustomRecordDelimiter = false;
        this.charset = null;
        this.userSettings = new UserSettings();
        this.initialized = false;
        this.closed = false;
        if (fileName == null) {
            throw new IllegalArgumentException("Parameter fileName can not be null.");
        }
        if (charset == null) {
            throw new IllegalArgumentException("Parameter charset can not be null.");
        }
        this.fileName = fileName;
        this.userSettings.Delimiter = delimiter;
        this.charset = charset;
    }
    
    public CsvWriter(final String fileName) {
        this(fileName, ',', Charset.forName("ISO-8859-1"));
    }
    
    public CsvWriter(final Writer outputStream, final char delimiter) {
        this.outputStream = null;
        this.fileName = null;
        this.firstColumn = true;
        this.useCustomRecordDelimiter = false;
        this.charset = null;
        this.userSettings = new UserSettings();
        this.initialized = false;
        this.closed = false;
        if (outputStream == null) {
            throw new IllegalArgumentException("Parameter outputStream can not be null.");
        }
        this.outputStream = new PrintWriter(outputStream);
        this.userSettings.Delimiter = delimiter;
        this.initialized = true;
    }
    
    public CsvWriter(final OutputStream outputStream, final char delimiter, final Charset charset) {
        this(new OutputStreamWriter(outputStream, charset), delimiter);
    }
    
    public char getDelimiter() {
        return this.userSettings.Delimiter;
    }
    
    public void setDelimiter(final char delimiter) {
        this.userSettings.Delimiter = delimiter;
    }
    
    public char getRecordDelimiter() {
        return this.userSettings.RecordDelimiter;
    }
    
    public void setRecordDelimiter(final char recordDelimiter) {
        this.useCustomRecordDelimiter = true;
        this.userSettings.RecordDelimiter = recordDelimiter;
    }
    
    public char getTextQualifier() {
        return this.userSettings.TextQualifier;
    }
    
    public void setTextQualifier(final char textQualifier) {
        this.userSettings.TextQualifier = textQualifier;
    }
    
    public boolean getUseTextQualifier() {
        return this.userSettings.UseTextQualifier;
    }
    
    public void setUseTextQualifier(final boolean useTextQualifier) {
        this.userSettings.UseTextQualifier = useTextQualifier;
    }
    
    public int getEscapeMode() {
        return this.userSettings.EscapeMode;
    }
    
    public void setEscapeMode(final int escapeMode) {
        this.userSettings.EscapeMode = escapeMode;
    }
    
    public void setComment(final char comment) {
        this.userSettings.Comment = comment;
    }
    
    public char getComment() {
        return this.userSettings.Comment;
    }
    
    public boolean getForceQualifier() {
        return this.userSettings.ForceQualifier;
    }
    
    public void setForceQualifier(final boolean forceQualifier) {
        this.userSettings.ForceQualifier = forceQualifier;
    }
    
    public void write(String content, final boolean preserveSpaces) throws IOException {
        this.checkClosed();
        this.checkInit();
        if (content == null) {
            content = "";
        }
        if (!this.firstColumn) {
            this.outputStream.write(this.userSettings.Delimiter);
        }
        boolean textQualify = this.userSettings.ForceQualifier;
        if (!preserveSpaces && content.length() > 0) {
            content = content.trim();
        }
        if (!textQualify && this.userSettings.UseTextQualifier && (content.indexOf(this.userSettings.TextQualifier) > -1 || content.indexOf(this.userSettings.Delimiter) > -1 || (!this.useCustomRecordDelimiter && (content.indexOf(10) > -1 || content.indexOf(13) > -1)) || (this.useCustomRecordDelimiter && content.indexOf(this.userSettings.RecordDelimiter) > -1) || (this.firstColumn && content.length() > 0 && content.charAt(0) == this.userSettings.Comment) || (this.firstColumn && content.length() == 0))) {
            textQualify = true;
        }
        if (this.userSettings.UseTextQualifier && !textQualify && content.length() > 0 && preserveSpaces) {
            final char firstLetter = content.charAt(0);
            if (firstLetter == ' ' || firstLetter == '\t') {
                textQualify = true;
            }
            if (!textQualify && content.length() > 1) {
                final char lastLetter = content.charAt(content.length() - 1);
                if (lastLetter == ' ' || lastLetter == '\t') {
                    textQualify = true;
                }
            }
        }
        if (textQualify) {
            this.outputStream.write(this.userSettings.TextQualifier);
            if (this.userSettings.EscapeMode == 2) {
                content = replace(content, "\\", "\\\\");
                content = replace(content, new StringBuilder().append(this.userSettings.TextQualifier).toString(), "\\" + this.userSettings.TextQualifier);
            }
            else {
                content = replace(content, new StringBuilder().append(this.userSettings.TextQualifier).toString(), new StringBuilder().append(this.userSettings.TextQualifier).append(this.userSettings.TextQualifier).toString());
            }
        }
        else if (this.userSettings.EscapeMode == 2) {
            content = replace(content, "\\", "\\\\");
            content = replace(content, new StringBuilder().append(this.userSettings.Delimiter).toString(), "\\" + this.userSettings.Delimiter);
            if (this.useCustomRecordDelimiter) {
                content = replace(content, new StringBuilder().append(this.userSettings.RecordDelimiter).toString(), "\\" + this.userSettings.RecordDelimiter);
            }
            else {
                content = replace(content, "\r", "\\\r");
                content = replace(content, "\n", "\\\n");
            }
            if (this.firstColumn && content.length() > 0 && content.charAt(0) == this.userSettings.Comment) {
                if (content.length() > 1) {
                    content = "\\" + this.userSettings.Comment + content.substring(1);
                }
                else {
                    content = "\\" + this.userSettings.Comment;
                }
            }
        }
        this.outputStream.write(content);
        if (textQualify) {
            this.outputStream.write(this.userSettings.TextQualifier);
        }
        this.firstColumn = false;
    }
    
    public void write(final String content) throws IOException {
        this.write(content, false);
    }
    
    public void writeComment(final String commentText) throws IOException {
        this.checkClosed();
        this.checkInit();
        this.outputStream.write(this.userSettings.Comment);
        this.outputStream.write(commentText);
        if (this.useCustomRecordDelimiter) {
            this.outputStream.write(this.userSettings.RecordDelimiter);
        }
        else {
            this.outputStream.println();
        }
        this.firstColumn = true;
    }
    
    public void writeRecord(final String[] values, final boolean preserveSpaces) throws IOException {
        if (values != null && values.length > 0) {
            for (int i = 0; i < values.length; ++i) {
                this.write(values[i], preserveSpaces);
            }
            this.endRecord();
        }
    }
    
    public void writeRecord(final String[] values) throws IOException {
        this.writeRecord(values, false);
    }
    
    public void endRecord() throws IOException {
        this.checkClosed();
        this.checkInit();
        if (this.useCustomRecordDelimiter) {
            this.outputStream.write(this.userSettings.RecordDelimiter);
        }
        else {
            this.outputStream.println();
        }
        this.firstColumn = true;
    }
    
    private void checkInit() throws IOException {
        if (!this.initialized) {
            if (this.fileName != null) {
                this.outputStream = new PrintWriter(new OutputStreamWriter(new FileOutputStream(this.fileName), this.charset));
            }
            this.initialized = true;
        }
    }
    
    public void flush() {
        this.outputStream.flush();
    }
    
    public void close() {
        if (!this.closed) {
            this.close(true);
            this.closed = true;
        }
    }
    
    private void close(final boolean closing) {
        if (!this.closed) {
            if (closing) {
                this.charset = null;
            }
            try {
                if (this.initialized) {
                    this.outputStream.close();
                }
            }
            catch (final Exception ex) {}
            this.outputStream = null;
            this.closed = true;
        }
    }
    
    private void checkClosed() throws IOException {
        if (this.closed) {
            throw new IOException("This instance of the CsvWriter class has already been closed.");
        }
    }
    
    @Override
    protected void finalize() {
        this.close(false);
    }
    
    public static String replace(final String original, final String pattern, final String replace) {
        final int len = pattern.length();
        int found = original.indexOf(pattern);
        if (found > -1) {
            final StringBuffer sb = new StringBuffer();
            int start;
            for (start = 0; found != -1; found = original.indexOf(pattern, start)) {
                sb.append(original.substring(start, found));
                sb.append(replace);
                start = found + len;
            }
            sb.append(original.substring(start));
            return sb.toString();
        }
        return original;
    }
    
    private class Letters
    {
        public static final char LF = '\n';
        public static final char CR = '\r';
        public static final char QUOTE = '\"';
        public static final char COMMA = ',';
        public static final char SPACE = ' ';
        public static final char TAB = '\t';
        public static final char POUND = '#';
        public static final char BACKSLASH = '\\';
        public static final char NULL = '\0';
    }
    
    private class UserSettings
    {
        public char TextQualifier;
        public boolean UseTextQualifier;
        public char Delimiter;
        public char RecordDelimiter;
        public char Comment;
        public int EscapeMode;
        public boolean ForceQualifier;
        
        public UserSettings() {
            this.TextQualifier = '\"';
            this.UseTextQualifier = true;
            this.Delimiter = ',';
            this.RecordDelimiter = '\0';
            this.Comment = '#';
            this.EscapeMode = 1;
            this.ForceQualifier = false;
        }
    }
}
